
import json, time
import datetime
import asyncio
from mqttlogic import AlarmConfig, PendingAlarm

class Alarm(object):

    def __init__(self):
        self.alarmConfigs = {}
        self.pendingAlarms = [] #PendingAlarm list
        #TODO save timestamp of last shutdown - so that we do not process sensor msgs again BEFORE that time on startup
        #or create alarm ID based on hash of sensor triggers (timestamp, reedstatus etc.) so that duplicates are detected

        # TODO on startup we first receive sensor msgs, and THEN the alarm configuration -> alarms of past sensor msgs do not fire

    def createAlarmJson(self, sensorId, msg):
        millis = int(round(time.time() * 1000))

        #alarmId = millis since 01.01.2020
        epoch2020 = datetime.datetime.now() - datetime.datetime.utcfromtimestamp(1577836800)
        alarmId = int(epoch2020.total_seconds() * 1000)

        jsonOutput = {}
        jsonOutput['sid'] = sensorId
        jsonOutput['aid'] = alarmId
        jsonOutput['ts'] = millis # unix timestamp in ms
        jsonOutput['msg'] = msg

        return json.dumps(jsonOutput)


    def createBatteryAlarmJson(self, sensorId, milliVolt, percentageLeft):
        millis = int(round(time.time() * 1000))

        alarmObj = {}
        alarmObj['event'] = 'battery_low'
        alarmObj['sid'] = int(sensorId)
        alarmObj['ts'] = millis # unix timestamp in ms
        alarmObj['mV'] = int(milliVolt)
        alarmObj['max_percentage_left'] = percentageLeft

        return json.dumps(alarmObj)


    def registerAlarmTrigger(self, alarmConfig):
        self.alarmConfigs[alarmConfig.uuid] = alarmConfig

    def unregisterAlarmTrigger(self, alarmConfig):
        self.alarmConfigs.pop(alarmConfig.uuid, None)

    def executeAlarmImmediately(self, alarmConfig, client, sensorId):
        print ('Alarm with config UUID {} triggered.'.format(alarmConfig.uuid))
        print ('Pending alarms: {}'.format(len(self.pendingAlarms)))

        alarmJson = self.createAlarmJson(sensorId, '')
        print (alarmJson)

        client.publish('alarms/clients/' + alarmConfig.clientId, alarmJson, qos=1, retain=False, content_type='json')

    async def executeAlarm(self, delay, pendingAlarm, client, sensorId):
        if delay > 0:
            try:
                await asyncio.sleep(delay) #TODO should not be a delay, but an absolute time: sensor reading time + alarm config delay
            except asyncio.CancelledError:
                print ('Canceled alarm with config uid {}'.format(pendingAlarm.alarmConfig.uuid))
                self.pendingAlarms.remove(pendingAlarm)
                print ('Pending alarms: {}'.format(len(self.pendingAlarms)))

        #delete task currently being executed
        self.pendingAlarms.remove(pendingAlarm)

        self.executeAlarmImmediately(pendingAlarm.alarmConfig, client, sensorId)


    def checkAlarmCondition(self, client, mqttPath, sensorId, sensorData, timestamp, properties):
        #unprocessed msgs would be received with properties['retain'] = 0, only @startup we get 1 msg with retain=1

        for pendingAlarm in self.pendingAlarms:
            #if time match + sensor ID match but no reed status match -> delete any tasks matching time + sensor ID
            pendingAlarm.cancelAlarmIfNoMatch(mqttPath, sensorData, timestamp)

        for uuid, alarmConfig in self.alarmConfigs.items():
            triggerDelay = alarmConfig.getTriggerDelay(mqttPath, sensorData, timestamp)

            if triggerDelay is None:
                continue

            if triggerDelay == 0:
                self.executeAlarmImmediately(alarmConfig, client, sensorId)
            elif triggerDelay > 0:
                print ('Alarm with config UUID {} scheduled to run in {} seconds'.format(uuid, triggerDelay))

                pAlarm = PendingAlarm.PendingAlarm(mqttPath, alarmConfig)
                asyncioTask = asyncio.create_task(self.executeAlarm(triggerDelay, pAlarm, client, sensorId))
                pAlarm.setTask(asyncioTask)
                self.pendingAlarms.append(pAlarm)
                print('Pending alarms: {}'.format(len(self.pendingAlarms)))
